@@ -9,7 +9,14 @@
 #include <sys/vfs.h>
 #include <sys/prctl.h>
 #include <ctype.h>
+#include <sys/sendfile.h>
 #include <sched.h>
+#include <sys/capability.h>
+
+#ifndef SEEK_DATA
+#define SEEK_DATA	3
+#define SEEK_HOLE	4
+#endif
 
 /* Stolen from kernel/fs/nfs/unlink.c */
 #define SILLYNAME_PREF ".nfs"
@@ -20,16 +27,17 @@
 #include "file-ids.h"
 #include "mount.h"
 #include "files.h"
-#include "list.h"
+#include "common/list.h"
 #include "rst-malloc.h"
 #include "fs-magic.h"
-#include "asm/atomic.h"
 #include "namespaces.h"
 #include "proc_parse.h"
 #include "pstree.h"
 #include "fault-injection.h"
+#include "external.h"
 
 #include "protobuf.h"
+#include "util.h"
 #include "images/regfile.pb-c.h"
 #include "images/remap-file-path.pb-c.h"
 
@@ -37,6 +45,7 @@
 #include "plugin.h"
 
 int setfsuid(uid_t fsuid);
+int setfsgid(gid_t fsuid);
 
 /*
  * Ghost files are those not visible from the FS. Dumping them is
@@ -56,7 +65,23 @@ struct ghost_file {
 static u32 ghost_file_ids = 1;
 static LIST_HEAD(ghost_files);
 
-static mutex_t *ghost_file_mutex;
+/*
+ * When opening remaps we first create a link on the remap
+ * target, then open one, then unlink. In case the remap
+ * source has more than one instance, these tree steps
+ * should be serialized with each other.
+ */
+static mutex_t *remap_open_lock;
+
+static inline int init_remap_lock(void)
+{
+	remap_open_lock = shmalloc(sizeof(*remap_open_lock));
+	if (!remap_open_lock)
+		return -1;
+
+	mutex_init(remap_open_lock);
+	return 0;
+}
 
 static LIST_HEAD(remaps);
 
@@ -80,7 +105,7 @@ static int note_link_remap(char *path, struct ns_id *nsid)
 	if (!rlb)
 		goto err;
 
-	rlb->path = strdup(path);
+	rlb->path = xstrdup(path);
 	if (!rlb->path)
 		goto err2;
 
@@ -123,15 +148,132 @@ static int trim_last_parent(char *path)
 	return 0;
 }
 
-static int mkreg_ghost(char *path, u32 mode, struct ghost_file *gf, struct cr_img *img)
+#define BUFSIZE	(4096)
+
+static int copy_chunk_from_file(int fd, int img, off_t off, size_t len)
+{
+	char *buf = NULL;
+	int ret;
+
+	while (len > 0) {
+		ret = sendfile(img, fd, &off, len);
+		if (ret <= 0) {
+			pr_perror("Can't send ghost to image");
+			return -1;
+		}
+
+		len -= ret;
+	}
+
+	xfree(buf);
+
+	return 0;
+}
+
+static int copy_file_to_chunks(int fd, struct cr_img *img, size_t file_size)
+{
+	GhostChunkEntry ce = GHOST_CHUNK_ENTRY__INIT;
+	off_t data, hole = 0;
+
+	while (hole < file_size) {
+		data = lseek(fd, hole, SEEK_DATA);
+		if (data < 0) {
+			if (errno == ENXIO)
+				/* No data */
+				break;
+			else if (hole == 0) {
+				/* No SEEK_HOLE/DATA by FS */
+				data = 0;
+				hole = file_size;
+			} else {
+				pr_perror("Can't seek file data");
+				return -1;
+			}
+		} else {
+			hole = lseek(fd, data, SEEK_HOLE);
+			if (hole < 0) {
+				pr_perror("Can't seek file hole");
+				return -1;
+			}
+		}
+
+		ce.len = hole - data;
+		ce.off = data;
+
+		if (pb_write_one(img, &ce, PB_GHOST_CHUNK))
+			return -1;
+
+		if (copy_chunk_from_file(fd, img_raw_fd(img), ce.off, ce.len))
+			return -1;
+	}
+
+	return 0;
+}
+
+static int copy_chunk_to_file(int img, int fd, off_t off, size_t len)
+{
+	char *buf = NULL;
+	int ret;
+
+	while (len > 0) {
+		if (lseek(fd, off, SEEK_SET) < 0) {
+			pr_perror("Can't seek file");
+			return -1;
+		}
+		ret = sendfile(fd, img, NULL, len);
+		if (ret < 0) {
+			pr_perror("Can't send data");
+			return -1;
+		}
+
+		off += ret;
+		len -= ret;
+	}
+
+	xfree(buf);
+
+	return 0;
+}
+
+static int copy_file_from_chunks(struct cr_img *img, int fd, size_t file_size)
+{
+	if (ftruncate(fd, file_size) < 0) {
+		pr_perror("Can't make file size");
+		return -1;
+	}
+
+	while (1) {
+		int ret;
+		GhostChunkEntry *ce;
+
+		ret = pb_read_one_eof(img, &ce, PB_GHOST_CHUNK);
+		if (ret <= 0)
+			return ret;
+
+		if (copy_chunk_to_file(img_raw_fd(img), fd, ce->off, ce->len))
+			return -1;
+
+		ghost_chunk_entry__free_unpacked(ce, NULL);
+	}
+}
+
+static int mkreg_ghost(char *path, GhostFileEntry *gfe, struct cr_img *img)
 {
 	int gfd, ret;
 
-	gfd = open(path, O_WRONLY | O_CREAT | O_EXCL, mode);
+	gfd = open(path, O_WRONLY | O_CREAT | O_EXCL, gfe->mode);
 	if (gfd < 0)
 		return -1;
 
-	ret = copy_file(img_raw_fd(img), gfd, 0);
+	if (gfe->chunks) {
+		if (!gfe->has_size) {
+			pr_err("Corrupted ghost image -> no size\n");
+			return -1;
+		}
+
+		ret = copy_file_from_chunks(img, gfd, gfe->size);
+	} else
+		ret = copy_file(img_raw_fd(img), gfd, 0);
 	if (ret < 0)
 		unlink(path);
 	close(gfd);
@@ -196,12 +338,10 @@ again:
 		if ((ret = mknod(path, gfe->mode, gfe->rdev)) < 0)
 			msg = "Can't create node for ghost dev";
 	} else if (S_ISDIR(gfe->mode)) {
-		if ((ret = mkdir(path, gfe->mode)) < 0) {
-			pr_perror("Can't make ghost dir");
-			goto err;
-		}
+		if ((ret = mkdirpat(AT_FDCWD, path, gfe->mode)) < 0)
+			msg = "Can't make ghost dir";
 	} else {
-		if ((ret = mkreg_ghost(path, gfe->mode, gf, img)) < 0)
+		if ((ret = mkreg_ghost(path, gfe, img)) < 0)
 			msg = "Can't create ghost regfile";
 	}
 
@@ -237,12 +377,10 @@ static inline void ghost_path(char *path, int plen,
 	snprintf(path, plen, "%s.cr.%x.ghost", rfi->path, rfe->remap_id);
 }
 
-static int open_remap_ghost(struct reg_file_info *rfi,
+static int collect_remap_ghost(struct reg_file_info *rfi,
 		RemapFilePathEntry *rfe)
 {
 	struct ghost_file *gf;
-	GhostFileEntry *gfe = NULL;
-	struct cr_img *img;
 
 	list_for_each_entry(gf, &ghost_files, list)
 		if (gf->id == rfe->remap_id)
@@ -260,9 +398,35 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 	if (!gf)
 		return -1;
 
-	gf->remap.rpath = xmalloc(PATH_MAX);
+	/*
+	 * The rpath is shmalloc-ed because we create the ghost
+	 * file in root task context and generate its path there.
+	 * However the path should be visible by the criu task
+	 * in order to remove the ghost files from root FS (see
+	 * try_clean_remaps()).
+	 */
+	gf->remap.rpath = shmalloc(PATH_MAX);
 	if (!gf->remap.rpath)
-		goto err;
+		return -1;
+	gf->remap.rpath[0] = 0;
+	gf->id = rfe->remap_id;
+	list_add_tail(&gf->list, &ghost_files);
+
+gf_found:
+	rfi->is_dir = gf->remap.is_dir;
+	rfi->remap = &gf->remap;
+	return 0;
+}
+
+static int open_remap_ghost(struct reg_file_info *rfi,
+					RemapFilePathEntry *rfe)
+{
+	struct ghost_file *gf = container_of(rfi->remap, struct ghost_file, remap);
+	GhostFileEntry *gfe = NULL;
+	struct cr_img *img;
+
+	if (rfi->remap->rpath[0])
+		return 0;
 
 	img = open_image(CR_FD_GHOST_FILE, O_RSTR, rfe->remap_id);
 	if (!img)
@@ -288,17 +452,13 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 	if (create_ghost(gf, gfe, img))
 		goto close_ifd;
 
-	ghost_file_entry__free_unpacked(gfe, NULL);
 	close_image(img);
 
-	gf->id = rfe->remap_id;
-	gf->remap.users = 0;
 	gf->remap.is_dir = S_ISDIR(gfe->mode);
-	gf->remap.owner = gfe->uid;
-	list_add_tail(&gf->list, &ghost_files);
-gf_found:
-	rfi->is_dir = gf->remap.is_dir;
-	rfi->remap = &gf->remap;
+	gf->remap.uid = gfe->uid;
+	gf->remap.gid = gfe->gid;
+	ghost_file_entry__free_unpacked(gfe, NULL);
+
 	return 0;
 
 close_ifd:
@@ -311,13 +471,12 @@ err:
 	return -1;
 }
 
-static int open_remap_linked(struct reg_file_info *rfi,
+static int collect_remap_linked(struct reg_file_info *rfi,
 		RemapFilePathEntry *rfe)
 {
 	struct file_remap *rm;
 	struct file_desc *rdesc;
 	struct reg_file_info *rrfi;
-	uid_t owner = -1;
 
 	rdesc = find_file_desc_raw(FD_TYPES__REG, rfe->remap_id);
 	if (!rdesc) {
@@ -332,30 +491,36 @@ static int open_remap_linked(struct reg_file_info *rfi,
 	rrfi = container_of(rdesc, struct reg_file_info, d);
 	pr_info("Remapped %s -> %s\n", rfi->path, rrfi->path);
 
-	if (root_ns_mask & CLONE_NEWUSER) {
-		int rfd;
-		struct stat st;
-
-		rfd = mntns_get_root_by_mnt_id(rfi->rfe->mnt_id);
-		if (fstatat(rfd, rrfi->path, &st, AT_SYMLINK_NOFOLLOW)) {
-			pr_perror("Can't get owner of link remap %s", rrfi->path);
-			xfree(rm);
-			return -1;
-		}
-
-		owner = st.st_uid;
-	}
-
 	rm->rpath = rrfi->path;
-	rm->users = 0;
 	rm->is_dir = false;
-	rm->owner = owner;
+	rm->uid = -1;
+	rm->gid = -1;
 	rm->rmnt_id = rfi->rfe->mnt_id;
 	rfi->remap = rm;
 	return 0;
 }
 
-static int open_remap_dead_process(struct reg_file_info *rfi,
+static int open_remap_linked(struct reg_file_info *rfi,
+		RemapFilePathEntry *rfe)
+{
+	if (root_ns_mask & CLONE_NEWUSER) {
+		int rfd;
+		struct stat st;
+
+		rfd = mntns_get_root_by_mnt_id(rfi->rfe->mnt_id);
+		if (fstatat(rfd, rfi->remap->rpath, &st, AT_SYMLINK_NOFOLLOW)) {
+			pr_perror("Can't get owner of link remap %s", rfi->remap->rpath);
+			return -1;
+		}
+
+		rfi->remap->uid = st.st_uid;
+		rfi->remap->gid = st.st_gid;
+	}
+
+	return 0;
+}
+
+static int collect_remap_dead_process(struct reg_file_info *rfi,
 		RemapFilePathEntry *rfe)
 {
 	struct pstree_item *helper;
@@ -364,7 +529,7 @@ static int open_remap_dead_process(struct reg_file_info *rfi,
 	if (!helper)
 		return -1;
 
-	if (helper->pid.state != TASK_UNDEF) {
+	if (helper->pid->state != TASK_UNDEF) {
 		pr_info("Skipping helper for restoring /proc/%d; pid exists\n", rfe->remap_id);
 		return 0;
 	}
@@ -373,11 +538,12 @@ static int open_remap_dead_process(struct reg_file_info *rfi,
 
 	helper->sid = root_item->sid;
 	helper->pgid = root_item->pgid;
-	helper->pid.virt = rfe->remap_id;
+	helper->pid->ns[0].virt = rfe->remap_id;
 	helper->parent = root_item;
+	helper->ids = root_item->ids;
 	list_add_tail(&helper->sibling, &root_item->children);
 
-	pr_info("Added a helper for restoring /proc/%d\n", helper->pid.virt);
+	pr_info("Added a helper for restoring /proc/%d\n", vpid(helper));
 
 	return 0;
 }
@@ -414,6 +580,23 @@ static int collect_one_remap(void *obj, ProtobufCMessage *msg, struct cr_img *i)
 
 	ri->rfi = container_of(fdesc, struct reg_file_info, d);
 
+	switch (rfe->remap_type) {
+	case REMAP_TYPE__GHOST:
+		if (collect_remap_ghost(ri->rfi, ri->rfe))
+			return -1;
+		break;
+	case REMAP_TYPE__LINKED:
+		if (collect_remap_linked(ri->rfi, ri->rfe))
+			return -1;
+		break;
+	case REMAP_TYPE__PROCFS:
+		if (collect_remap_dead_process(ri->rfi, rfe) < 0)
+			return -1;
+		break;
+	default:
+		break;
+	}
+
 	list_add_tail(&ri->list, &remaps);
 
 	return 0;
@@ -435,7 +618,7 @@ static int prepare_one_remap(struct remap_info *ri)
 		ret = open_remap_ghost(rfi, rfe);
 		break;
 	case REMAP_TYPE__PROCFS:
-		/* handled earlier by prepare_procfs_remaps */
+		/* handled earlier by collect_remap_dead_process */
 		ret = 0;
 		break;
 	default:
@@ -447,36 +630,15 @@ out:
 	return ret;
 }
 
-/* We separate the preparation of PROCFS remaps because they allocate pstree
- * items, which need to be seen by the root task. We can't do all remaps here,
- * because the files haven't been loaded yet.
- */
-int prepare_procfs_remaps(void)
-{
-	struct remap_info *ri;
-
-	list_for_each_entry(ri, &remaps, list) {
-		RemapFilePathEntry *rfe = ri->rfe;
-		struct reg_file_info *rfi = ri->rfi;
-
-		switch (rfe->remap_type) {
-		case REMAP_TYPE__PROCFS:
-			if (open_remap_dead_process(rfi, rfe) < 0)
-				return -1;
-			break;
-		default:
-			continue;
-		}
-	}
-
-	return 0;
-}
-
 int prepare_remaps(void)
 {
 	struct remap_info *ri;
 	int ret = 0;
 
+	ret = init_remap_lock();
+	if (ret)
+		return ret;
+
 	list_for_each_entry(ri, &remaps, list) {
 		ret = prepare_one_remap(ri);
 		if (ret)
@@ -486,73 +648,59 @@ int prepare_remaps(void)
 	return ret;
 }
 
-static void try_clean_ghost(struct remap_info *ri)
+static int clean_one_remap(struct remap_info *ri)
 {
 	char path[PATH_MAX];
-	int mnt_id, ret;
+	int mnt_id, ret, rmntns_root;
+	struct file_remap *remap = ri->rfi->remap;
+
+	if (remap->rpath[0] == 0)
+		return 0;
 
 	mnt_id = ri->rfi->rfe->mnt_id; /* rirfirfe %) */
 	ret = rst_get_mnt_root(mnt_id, path, sizeof(path));
 	if (ret < 0)
-		return;
+		return -1;
 	if (ret >= sizeof(path) - 1) {
 		pr_err("The path buffer is too small\n");
-		return;
-	}
-	if (path[ret] != '/') {
-		path[ret++] = '/';
-		path[ret] = 0;
-	}
-
-	if (ri->rfi->remap == NULL)
-		return;
-	if (!ri->rfi->is_dir) {
-		ghost_path(path + ret, sizeof(path) - ret, ri->rfi, ri->rfe);
-		if (!unlink(path)) {
-			pr_info(" `- X [%s] ghost\n", path);
-			return;
-		}
-	} else {
-		strncpy(path + ret, ri->rfi->path, sizeof(path) - ret);
-		if (!rmdir(path)) {
-			pr_info(" `- Xd [%s] ghost\n", path);
-			return;
-		}
-	}
-
-	pr_perror(" `- XFail [%s] ghost", path);
-}
-
-static int clean_one_remap(struct file_remap *remap)
-{
-	int rmntns_root, ret = 0;
-
-	rmntns_root = mntns_get_root_by_mnt_id(remap->rmnt_id);
-	if (rmntns_root < 0)
 		return -1;
+	}
+
+	rmntns_root = open(path, O_RDONLY);
+	if (rmntns_root < 0) {
+		pr_perror("Unbale to open %s", path);
+		return -1;
+	}
 
 	pr_info("Unlink remap %s\n", remap->rpath);
 
 	ret = unlinkat(rmntns_root, remap->rpath, remap->is_dir ? AT_REMOVEDIR : 0);
 	if (ret < 0) {
-		pr_perror("Couldn't unlink remap %s", remap->rpath);
+		close(rmntns_root);
+		pr_perror("Couldn't unlink remap %d %s", rmntns_root, remap->rpath);
 		return -1;
 	}
+	close(rmntns_root);
+	remap->rpath[0] = 0;
 
 	return 0;
 }
 
-void try_clean_remaps()
+int try_clean_remaps(bool only_ghosts)
 {
 	struct remap_info *ri;
+	int ret = 0;
 
-	if (list_empty(&remaps))
-		return;
-
-	list_for_each_entry(ri, &remaps, list)
+	list_for_each_entry(ri, &remaps, list) {
 		if (ri->rfe->remap_type == REMAP_TYPE__GHOST)
-			try_clean_ghost(ri);
+			ret |= clean_one_remap(ri);
+		else if (only_ghosts)
+			continue;
+		else if (ri->rfe->remap_type == REMAP_TYPE__LINKED)
+			ret |= clean_one_remap(ri);
+	}
 
+	return ret;
 }
 
 static struct collect_image_info remap_cinfo = {
@@ -562,6 +710,9 @@ static struct collect_image_info remap_cinfo = {
 	.collect = collect_one_remap,
 };
 
+/* Tiny files don't need to generate chunks in ghost image. */
+#define GHOST_CHUNKS_THRESH	(3 * 4096)
+
 static int dump_ghost_file(int _fd, u32 id, const struct stat *st, dev_t phys_dev)
 {
 	struct cr_img *img;
@@ -594,6 +745,12 @@ static int dump_ghost_file(int _fd, u32 id, const struct stat *st, dev_t phys_de
 		gfe.rdev = st->st_rdev;
 	}
 
+	if (S_ISREG(st->st_mode) && (st->st_size >= GHOST_CHUNKS_THRESH)) {
+		gfe.has_chunks = gfe.chunks = true;
+		gfe.has_size = true;
+		gfe.size = st->st_size;
+	}
+
 	if (pb_write_one(img, &gfe, PB_GHOST_FILE))
 		return -1;
 
@@ -611,7 +768,11 @@ static int dump_ghost_file(int _fd, u32 id, const struct stat *st, dev_t phys_de
 			pr_perror("Can't open ghost original file");
 			return -1;
 		}
-		ret = copy_file(fd, img_raw_fd(img), st->st_size);
+
+		if (gfe.chunks)
+			ret = copy_file_to_chunks(fd, img, st->st_size);
+		else
+			ret = copy_file(fd, img_raw_fd(img), st->st_size);
 		close(fd);
 		if (ret)
 			return -1;
@@ -621,27 +782,15 @@ static int dump_ghost_file(int _fd, u32 id, const struct stat *st, dev_t phys_de
 	return 0;
 }
 
-void remap_put(struct file_remap *remap)
-{
-	mutex_lock(ghost_file_mutex);
-	if (--remap->users == 0)
-		clean_one_remap(remap);
-	mutex_unlock(ghost_file_mutex);
-}
-
 struct file_remap *lookup_ghost_remap(u32 dev, u32 ino)
 {
 	struct ghost_file *gf;
 
-	mutex_lock(ghost_file_mutex);
 	list_for_each_entry(gf, &ghost_files, list) {
 		if (gf->ino == ino && (gf->dev == dev)) {
-			gf->remap.users++;
-			mutex_unlock(ghost_file_mutex);
 			return &gf->remap;
 		}
 	}
-	mutex_unlock(ghost_file_mutex);
 
 	return NULL;
 }
@@ -710,9 +859,11 @@ static void __rollback_link_remaps(bool do_unlink)
 
 void delete_link_remaps(void) { __rollback_link_remaps(true); }
 void free_link_remaps(void) { __rollback_link_remaps(false); }
+static int linkat_hard(int odir, char *opath, int ndir, char *npath, uid_t uid, gid_t gid, int flags);
 
 static int create_link_remap(char *path, int len, int lfd,
-				u32 *idp, struct ns_id *nsid)
+				u32 *idp, struct ns_id *nsid,
+				const struct stat *st)
 {
 	char link_name[PATH_MAX], *tmp;
 	RegFileEntry rfe = REG_FILE_ENTRY__INIT;
@@ -756,7 +907,8 @@ static int create_link_remap(char *path, int len, int lfd,
 	mntns_root = mntns_get_root_fd(nsid);
 
 again:
-	ret = linkat(lfd, "", mntns_root, link_name, AT_EMPTY_PATH);
+	ret = linkat_hard(lfd, "", mntns_root, link_name,
+				st->st_uid, st->st_gid, AT_EMPTY_PATH);
 	if (ret < 0 && errno == ENOENT) {
 		/* Use grand parent, if parent directory does not exist. */
 		if (trim_last_parent(link_name) < 0) {
@@ -781,11 +933,13 @@ static int dump_linked_remap(char *path, int len, const struct stat *ost,
 	u32 lid;
 	RemapFilePathEntry rpe = REMAP_FILE_PATH_ENTRY__INIT;
 
-	if (create_link_remap(path, len, lfd, &lid, nsid))
+	if (create_link_remap(path, len, lfd, &lid, nsid, ost))
 		return -1;
 
 	rpe.orig_id = id;
 	rpe.remap_id = lid;
+	rpe.has_remap_type = true;
+	rpe.remap_type = REMAP_TYPE__LINKED;
 
 	return pb_write_one(img_from_set(glob_imgset, CR_FD_REMAP_FPATH),
 			&rpe, PB_REMAP_FPATH);
@@ -813,14 +967,14 @@ int dead_pid_conflict(void)
 			 * If the dead PID was given to a main thread of another
 			 * process, this is handled during restore.
 			 */
-			item = container_of(node, struct pstree_item, pid);
-			if (item->pid.real == item->threads[i].real ||
-			    item->threads[i].virt != pid)
+			item = node->item;
+			if (item->pid->real == item->threads[i].real ||
+			    item->threads[i].ns[0].virt != pid)
 				continue;
 		}
 
 		pr_err("Conflict with a dead task with the same PID as of this thread (virt %d, real %d).\n",
-			node->virt, node->real);
+			node->ns[0].virt, node->real);
 		return -1;
 	}
 
@@ -1189,16 +1343,18 @@ static void convert_path_from_another_mp(char *src, char *dst, int dlen,
 				src + off);
 }
 
-static int linkat_hard(int odir, char *opath, int ndir, char *npath, uid_t owner)
+static int linkat_hard(int odir, char *opath, int ndir, char *npath, uid_t uid, gid_t gid, int flags)
 {
-	int ret, old_fsuid = -1;
+	struct __user_cap_data_struct data[_LINUX_CAPABILITY_U32S_3];
+	struct __user_cap_header_struct hdr;
+	int ret, old_fsuid = -1, old_fsgid = -1;
 	int errno_save;
 
-	ret = linkat(odir, opath, ndir, npath, 0);
+	ret = linkat(odir, opath, ndir, npath, flags);
 	if (ret == 0)
 		return 0;
 
-	if (!( (errno == EPERM) && (root_ns_mask & CLONE_NEWUSER) )) {
+	if (!( (errno == EPERM || errno == EOVERFLOW) && (root_ns_mask & CLONE_NEWUSER) )) {
 		errno_save = errno;
 		pr_perror("Can't link %s -> %s", opath, npath);
 		errno = errno_save;
@@ -1219,16 +1375,42 @@ static int linkat_hard(int odir, char *opath, int ndir, char *npath, uid_t owner
 	 *
 	 * Fortunately, the setfsuid() requires ns-level
 	 * CAP_SETUID which we have.
+	 *
+	 * Starting with 4.8 the kernel doesn't allow to create inodes
+	 * with a uid or gid unknown to an user namespace.
+	 * 036d523641c66 ("vfs: Don't create inodes with a uid or gid unknown to the vfs")
 	 */
 
-	old_fsuid = setfsuid(owner);
+	old_fsuid = setfsuid(uid);
+	old_fsgid = setfsgid(gid);
 
-	ret = linkat(odir, opath, ndir, npath, 0);
+	/* AT_EMPTY_PATH requires CAP_DAC_READ_SEARCH */
+	if (flags & AT_EMPTY_PATH) {
+		hdr.version = _LINUX_CAPABILITY_VERSION_3;
+		hdr.pid = 0;
+
+		if (capget(&hdr, data) < 0) {
+			errno_save = errno;
+			pr_perror("capget");
+			goto out;
+		}
+		data[0].effective = data[0].permitted;
+		data[1].effective = data[1].permitted;
+		if (capset(&hdr, data) < 0) {
+			errno_save = errno;
+			pr_perror("capset");
+			goto out;
+		}
+	}
+
+	ret = linkat(odir, opath, ndir, npath, flags);
 	errno_save = errno;
 	if (ret < 0)
 		pr_perror("Can't link %s -> %s", opath, npath);
 
+out:
 	setfsuid(old_fsuid);
+	setfsgid(old_fsgid);
 	if (setfsuid(-1) != old_fsuid) {
 		pr_warn("Failed to restore old fsuid!\n");
 		/*
@@ -1258,7 +1440,8 @@ static void rm_parent_dirs(int mntns_root, char *path, int count)
 	if (!count)
 		return;
 
-	while (count--) {
+	while (count > 0) {
+		count -= 1;
 		p = strrchr(path, '/');
 		if (p)
 			*p = '\0';
@@ -1285,7 +1468,7 @@ static int make_parent_dirs_if_need(int mntns_root, char *path)
 
 	p = last_delim = strrchr(path, '/');
 	if (!p) {
-		pr_err("Path %s has no parent dir", path);
+		pr_err("Path %s has no parent dir\n", path);
 		return -1;
 	}
 	*p = '\0';
@@ -1379,8 +1562,11 @@ out_root:
 	if (*level < 0)
 		return -1;
 
-	if (linkat_hard(mntns_root, rpath, mntns_root, path, rfi->remap->owner) < 0) {
+	if (linkat_hard(mntns_root, rpath, mntns_root, path,
+			rfi->remap->uid, rfi->remap->gid, 0) < 0) {
+		int errno_saved = errno;
 		rm_parent_dirs(mntns_root, path, *level);
+		errno = errno_saved;
 		return -1;
 	}
 
@@ -1390,7 +1576,7 @@ out_root:
 int open_path(struct file_desc *d,
 		int(*open_cb)(int mntns_root, struct reg_file_info *, void *), void *arg)
 {
-	int tmp, mntns_root, level;
+	int tmp, mntns_root, level = 0;
 	struct reg_file_info *rfi;
 	char *orig_path = NULL;
 	char path[PATH_MAX];
@@ -1417,7 +1603,7 @@ int open_path(struct file_desc *d,
 			BUG();
 		}
 
-		mutex_lock(ghost_file_mutex);
+		mutex_lock(remap_open_lock);
 		if (rfi->remap->is_dir) {
 			/*
 			 * FIXME Can't make directory under new name.
@@ -1503,11 +1689,7 @@ ext:
 			rm_parent_dirs(mntns_root, rfi->path, level);
 		}
 
-		BUG_ON(!rfi->remap->users);
-		if (--rfi->remap->users == 0)
-			clean_one_remap(rfi->remap);
-
-		mutex_unlock(ghost_file_mutex);
+		mutex_unlock(remap_open_lock);
 	}
 	if (orig_path)
 		rfi->path = orig_path;
@@ -1579,6 +1761,60 @@ int open_reg_by_id(u32 id)
 	return open_reg_fd(fd);
 }
 
+struct filemap_ctx {
+	u32 flags;
+	struct file_desc *desc;
+	int fd;
+	/*
+	 * Whether or not to close the fd when we're about to
+	 * put a new one into ctx.
+	 *
+	 * True is used by premap, so that it just calls vm_open
+	 * in sequence, immediatelly mmap()s the file, then it
+	 * can be closed.
+	 *
+	 * False is used by open_vmas() which pre-opens the files
+	 * for restorer, and the latter mmap()s them and closes.
+	 *
+	 * ...
+	 */
+	bool close;
+	/* ...
+	 *
+	 * but closing all vmas won't work, as some of them share
+	 * the descriptor, so only the ones that terminate the
+	 * fd-sharing chain are marked with VMA_CLOSE flag, saying
+	 * restorer to close the vma's fd.
+	 *
+	 * Said that, this vma pointer references the previously
+	 * seen vma, so that once fd changes, this one gets the
+	 * closing flag.
+	 */
+	struct vma_area *vma;
+};
+
+static struct filemap_ctx ctx;
+
+void filemap_ctx_init(bool auto_close)
+{
+	ctx.desc = NULL;	/* to fail the first comparison in open_ */
+	ctx.fd = -1;		/* not to close random fd in _fini */
+	ctx.vma = NULL;		/* not to put spurious VMA_CLOSE in _fini */
+				/* flags may remain any */
+	ctx.close = auto_close;
+}
+
+void filemap_ctx_fini(void)
+{
+	if (ctx.close) {
+		if (ctx.fd >= 0)
+			close(ctx.fd);
+	} else {
+		if (ctx.vma)
+			ctx.vma->e->status |= VMA_CLOSE;
+	}
+}
+
 static int open_filemap(int pid, struct vma_area *vma)
 {
 	u32 flags;
@@ -1590,20 +1826,23 @@ static int open_filemap(int pid, struct vma_area *vma)
 	 * We open file w/o lseek, as mappings don't care about it
 	 */
 
-	BUG_ON(vma->vmfd == NULL);
-	if (vma->e->has_fdflags)
-		flags = vma->e->fdflags;
-	else if ((vma->e->prot & PROT_WRITE) &&
-			vma_area_is(vma, VMA_FILE_SHARED))
-		flags = O_RDWR;
-	else
-		flags = O_RDONLY;
+	BUG_ON((vma->vmfd == NULL) || !vma->e->has_fdflags);
+	flags = vma->e->fdflags;
 
-	ret = open_path(vma->vmfd, do_open_reg_noseek_flags, &flags);
-	if (ret < 0)
-		return ret;
+	if (ctx.flags != flags || ctx.desc != vma->vmfd) {
+		ret = open_path(vma->vmfd, do_open_reg_noseek_flags, &flags);
+		if (ret < 0)
+			return ret;
 
-	vma->e->fd = ret;
+		filemap_ctx_fini();
+
+		ctx.flags = flags;
+		ctx.desc = vma->vmfd;
+		ctx.fd = ret;
+	}
+
+	ctx.vma = vma;
+	vma->e->fd = ctx.fd;
 	return 0;
 }
 
@@ -1611,6 +1850,16 @@ int collect_filemap(struct vma_area *vma)
 {
 	struct file_desc *fd;
 
+	if (!vma->e->has_fdflags) {
+		/* Make a wild guess for the fdflags */
+		vma->e->has_fdflags = true;
+		if ((vma->e->prot & PROT_WRITE) &&
+				vma_area_is(vma, VMA_FILE_SHARED))
+			vma->e->fdflags = O_RDWR;
+		else
+			vma->e->fdflags = O_RDONLY;
+	}
+
 	fd = collect_special_file(vma->e->shmid);
 	if (!fd)
 		return -1;
@@ -1620,31 +1869,15 @@ int collect_filemap(struct vma_area *vma)
 	return 0;
 }
 
-static void remap_get(struct file_desc *fdesc, char typ)
+static int open_fe_fd(struct file_desc *fd, int *new_fd)
 {
-	struct reg_file_info *rfi;
+	int tmp;
 
-	rfi = container_of(fdesc, struct reg_file_info, d);
-	if (rfi->remap) {
-		pr_debug("One more remap user (%c) for %s\n",
-				typ, rfi->remap->rpath);
-		/* No lock, we're still sngle-process here */
-		rfi->remap->users++;
-	}
-}
-
-static void collect_reg_fd(struct file_desc *fdesc,
-		struct fdinfo_list_entry *fle, struct rst_info *ri)
-{
-	if (list_empty(&fdesc->fd_info_head))
-		remap_get(fdesc, 'f');
-
-	collect_gen_fd(fle, ri);
-}
-
-static int open_fe_fd(struct file_desc *fd)
-{
-	return open_path(fd, do_open_reg, NULL);
+	tmp = open_path(fd, do_open_reg, NULL);
+	if (tmp < 0)
+		return -1;
+	*new_fd = tmp;
+	return 0;
 }
 
 static char *reg_file_path(struct file_desc *d, char *buf, size_t s)
@@ -1658,7 +1891,6 @@ static char *reg_file_path(struct file_desc *d, char *buf, size_t s)
 static struct file_desc_ops reg_desc_ops = {
 	.type = FD_TYPES__REG,
 	.open = open_fe_fd,
-	.collect_fd = collect_reg_fd,
 	.name = reg_file_path,
 };
 
@@ -1680,7 +1912,6 @@ struct file_desc *try_collect_special_file(u32 id, int optional)
 		return NULL;
 	}
 
-	remap_get(fdesc, 's');
 	return fdesc;
 }
 
@@ -1710,16 +1941,6 @@ static struct collect_image_info reg_file_cinfo = {
 	.flags = COLLECT_SHARED,
 };
 
-int prepare_shared_reg_files(void)
-{
-	ghost_file_mutex = shmalloc(sizeof(*ghost_file_mutex));
-	if (!ghost_file_mutex)
-		return -1;
-
-	mutex_init(ghost_file_mutex);
-	return 0;
-}
-
 int collect_remaps_and_regfiles(void)
 {
 	if (collect_image(&reg_file_cinfo))
